@@ -1,3 +1,5 @@
+import collections
+
 import numpy as np
 from numba import njit
 from scipy import stats
@@ -15,7 +17,7 @@ def from_bnlearn(dag):
     return output
 
 
-def score(dag, data, score_function='bic', alpha=0.3):
+def score(dag, data, score_function='bic'):
     if type(dag) is str:
         dag = from_bnlearn(dag)
     elif type(dag) is not dict:
@@ -31,14 +33,7 @@ def score(dag, data, score_function='bic', alpha=0.3):
     sco = 0
     for var in varnames:
         cols = [varnames.index(v) for v in [var] + dag[var]['par']]
-        if score_function in ['nal', 'nal_g']:
-            sco += globals()[score_function](data, np.asarray(cols), alpha)
-        else:
-            try:
-                sco += globals()[score_function](data, np.asarray(cols))
-            except Exception as e:
-                raise Exception('score function ' + str(
-                    e) + ' is undefined or does not fit to data type. Available score functions are: bic (BIC for discrete variables) and nal (NAL for discrete variables).')
+        sco += local_score(data, cols, score_function)
     return sco
 
 
@@ -78,13 +73,24 @@ def bic_counter(data, arities, cols, weight=None):
     return bic
 
 
-def nal(data, cols, alpha=0.3, mul=1):
+def nal(data, cols, weight=None):
     arities = np.amax(data, axis=0) + 1
     data_local = np.delete(data, data[:, cols].min(axis=1) < 0, axis=0)
+    alpha = 0.3 + data_local.shape[0] / data.shape[0] / 10
     if data_local.shape[0] == 0:
         return -np.inf
     else:
-        return nal_counter(data_local, arities, cols, data.shape[0], 0.3 + data_local.shape[0] / data.shape[0] / 10, mul)
+        return nal_counter(data_local, arities, cols,
+                           np.ones(data_local.shape[0]) * data.shape[0] / data_local.shape[0], alpha)
+
+
+def nal_ipw(data, cols, weight):
+    arities = np.amax(data, axis=0) + 1
+    alpha = 0.3 + data.shape[0] / sum(weight) / 10
+    if data.shape[0] == 0:
+        return -np.inf
+    else:
+        return nal_counter(data, arities, cols, weight, alpha)
 
 
 def nal_sev(data, cols, cols_rvs):
@@ -94,7 +100,9 @@ def nal_sev(data, cols, cols_rvs):
 
 
 @njit(fastmath=True)
-def nal_counter(data, arities, cols, datasize, alpha=0.3, mul=1):
+def nal_counter(data, arities, cols, weight=None, alpha=0.3):
+    if weight is None:
+        weight = np.ones(data.shape[0])
     strides = np.empty(len(cols), dtype=np.uint32)
     idx = len(cols) - 1
     stride = 1
@@ -117,18 +125,18 @@ def nal_counter(data, arities, cols, datasize, alpha=0.3, mul=1):
                 if i != 0:
                     idx_ij += data[rowidx, cols[i]] * strides[i]
         if not miss_flag:
-            N_ijk[idx_ijk] += 1
+            N_ijk[idx_ijk] += weight[rowidx]
             for i in range(arities[cols[0]]):
-                N_ij[idx_ij + i * strides[0]] += 1
+                N_ij[idx_ij + i * strides[0]] += weight[rowidx]
     nal = 0
     for i in range(stride):
         if N_ijk[i] != 0:
             nal += N_ijk[i] * np.log(N_ijk[i] / N_ij[i])
     nal /= np.sum(N_ijk)
     if alpha == 1:
-        nal -= 0.5 * np.log(datasize) / datasize * (arities[cols[0]] - 1) * strides[0]
+        nal -= 0.5 * np.log(np.sum(weight)) / np.sum(weight) * (arities[cols[0]] - 1) * strides[0]
     else:
-        nal -= mul / data.shape[1] * np.power(datasize, -alpha) * (arities[cols[0]] - 1) * strides[0]
+        nal -= 1 / data.shape[1] * np.power(np.sum(weight), -alpha) * (arities[cols[0]] - 1) * strides[0]
     return nal
 
 
@@ -196,14 +204,12 @@ def density_counter(data, arities):
     return CT
 
 
-def local_score(data, cols, score_function='default', weights=None, alpha=0.3, mul=1):
+def local_score(data, cols, score_function='default', weights=None):
     '''
     :param weights: weight for data
     :param data: numbered version of data set
     :param cols: the index of node and its parents, the first element represents the index of the node and the following elements represent the indices of its parents
     :param score_function: name of score function, currently support bic, nal, bic_g
-    :param alpha: alpha parameter for NAL score
-    :param mul: multiplicator used for NAL score
     :return: local score of node (cols[0]) given its parents (cols[1:])
     '''
     if len(data) == 0:
@@ -211,12 +217,9 @@ def local_score(data, cols, score_function='default', weights=None, alpha=0.3, m
     else:
         if score_function == 'default':
             score_function = 'bic' if 'int' in data.dtype.name else 'bic_g'
-        if score_function in ['nal', 'nal_g']:
-            ls = globals()[score_function](data, np.asarray(cols), alpha, mul)
-        else:
-            try:
-                ls = globals()[score_function](data, np.asarray(cols), weights)
-            except Exception as e:
-                raise Exception('score function ' + str(
-                    e) + ' is undefined or does not fit to data type. Available score functions are: bic (BIC for discrete variables) and nal (NAL for discrete variables).')
+        try:
+            ls = globals()[score_function](data, np.asarray(cols), weights)
+        except Exception as e:
+            raise Exception('score function ' + str(
+                e) + ' is undefined or does not fit to data type. Available score functions are: bic (BIC for discrete variables) and nal (NAL for discrete variables).')
         return ls
