@@ -64,7 +64,7 @@ def check_cycle(vi, vj, dag):
     return cyc_flag
 
 
-def hc(data, method='standard', score_function='default', alpha=0.3, mul=1, debug=False):
+def hc(data, method='standard', score_function='default', debug=False):
     '''
         :param data: the training data used for learn BN
         :param method: the method for dealing with missing values, including:
@@ -153,7 +153,7 @@ def hc(data, method='standard', score_function='default', alpha=0.3, mul=1, debu
                         if W not in cache_score[vj][par_cur]:
                             cols = [varnames.index(vj)] + [varnames.index(x) for x in varnames if x in par_cur]
                             cache_score[vj][par_cur][W] = local_score(cache_data[W], cols, score_function,
-                                                                      cache_weight[W], alpha=alpha)
+                                                                      cache_weight[W])
                         # compute the local score for the searching graph
                         par_sea = tuple(sorted(dag[vj]['par'] + [vi]))
                         if par_sea not in cache_score[vj]:
@@ -161,7 +161,7 @@ def hc(data, method='standard', score_function='default', alpha=0.3, mul=1, debu
                         if W not in cache_score[vj][par_sea]:
                             cols = [varnames.index(vj)] + [varnames.index(x) for x in varnames if x in par_sea]
                             cache_score[vj][par_sea][W] = local_score(cache_data[W], cols, score_function,
-                                                                      cache_weight[W], alpha=alpha)
+                                                                      cache_weight[W])
                         if cache_score[vj][par_sea][W] != np.nan:
                             diff_temp = cache_score[vj][par_sea][W] - cache_score[vj][par_cur][W]
                             if debug:
@@ -200,14 +200,14 @@ def hc(data, method='standard', score_function='default', alpha=0.3, mul=1, debu
                             cols = [varnames.index(par_vi)] + [varnames.index(x) for x in varnames if
                                                                x in par_cur_par_vi]
                             cache_score[par_vi][par_cur_par_vi][W] = local_score(cache_data[W], cols, score_function,
-                                                                                 cache_weight[W], alpha=alpha)
+                                                                                 cache_weight[W])
                         par_cur_vi = tuple(dag[vi]['par'])
                         if par_cur_vi not in cache_score[vi]:
                             cache_score[vi][par_cur_vi] = {}
                         if W not in cache_score[vi][par_cur_vi]:
                             cols = [varnames.index(vi)] + [varnames.index(x) for x in varnames if x in par_cur_vi]
                             cache_score[vi][par_cur_vi][W] = local_score(cache_data[W], cols, score_function,
-                                                                         cache_weight[W], alpha=alpha)
+                                                                         cache_weight[W])
                         # compute the local score for the searching graph
                         par_sea_par_vi = tuple(sorted(dag[par_vi]['par'] + [vi]))
                         if par_sea_par_vi not in cache_score[par_vi]:
@@ -216,15 +216,14 @@ def hc(data, method='standard', score_function='default', alpha=0.3, mul=1, debu
                             cols = [varnames.index(par_vi)] + [varnames.index(x) for x in varnames if
                                                                x in par_sea_par_vi]
                             cache_score[par_vi][par_sea_par_vi][W] = local_score(cache_data[W], cols,
-                                                                                 score_function, cache_weight[W],
-                                                                                 alpha=alpha)
+                                                                                 score_function, cache_weight[W])
                         par_sea_vi = tuple([x for x in dag[vi]['par'] if x != par_vi])
                         if par_sea_vi not in cache_score[vi]:
                             cache_score[vi][par_sea_vi] = {}
                         if W not in cache_score[vi][par_sea_vi]:
                             cols = [varnames.index(vi)] + [varnames.index(x) for x in varnames if x in par_sea_vi]
                             cache_score[vi][par_sea_vi][W] = local_score(cache_data[W], cols, score_function,
-                                                                         cache_weight[W], alpha=alpha)
+                                                                         cache_weight[W])
                         if cache_score[vi][par_cur_vi][W] != np.nan and cache_score[par_vi][par_sea_par_vi][
                             W] != np.nan:
                             diff_temp = cache_score[vi][par_sea_vi][W] + cache_score[par_vi][par_sea_par_vi][W] - \
@@ -257,7 +256,7 @@ def hc(data, method='standard', score_function='default', alpha=0.3, mul=1, debu
                         if W not in cache_score[vi][par_cur]:
                             cols = [varnames.index(vi)] + [varnames.index(x) for x in varnames if x in par_cur]
                             cache_score[vi][par_cur][W] = local_score(cache_data[W], cols, score_function,
-                                                                      cache_weight[W], alpha=alpha)
+                                                                      cache_weight[W])
                         # compute the local score for the searching graph
                         par_sea = tuple([x for x in dag[vi]['par'] if x != par_vi])
                         if par_sea not in cache_score[vi]:
@@ -265,7 +264,7 @@ def hc(data, method='standard', score_function='default', alpha=0.3, mul=1, debu
                         if W not in cache_score[vi][par_sea]:
                             cols = [varnames.index(vi)] + [varnames.index(x) for x in varnames if x in par_sea]
                             cache_score[vi][par_sea][W] = local_score(cache_data[W], cols, score_function,
-                                                                      cache_weight[W], alpha=alpha)
+                                                                      cache_weight[W])
                         if cache_score[vi][par_cur][W] != np.nan:
                             diff_temp = cache_score[vi][par_sea][W] - cache_score[vi][par_cur][W]
                             if diff_temp - diff > 1e-10:
@@ -303,8 +302,7 @@ def hc(data, method='standard', score_function='default', alpha=0.3, mul=1, debu
             raise Exception('Mixed data is not supported.')
         # initialise dag, cached scores and candidate parents for each variable
         dag = {var: {'par': [], 'nei': []} for var in varnames}
-        cache = {var: {tuple([]): local_score(data, [varnames.index(var)], score_function, alpha=alpha, mul=mul)} for
-                 var in varnames}
+        cache = {var: {tuple([]): local_score(data, [varnames.index(var)], score_function)} for var in varnames}
         candidate = {var: [v for v in varnames if v is not var] for var in varnames}
         time_preprocess = time.time() - start_time
         diff = 1
@@ -324,7 +322,7 @@ def hc(data, method='standard', score_function='default', alpha=0.3, mul=1, debu
                         par_sea = tuple(sorted(dag[vj]['par'] + [vi]))
                         if par_sea not in cache[vj]:
                             cols = [varnames.index(x) for x in (vj,) + par_sea]
-                            cache[vj][par_sea] = local_score(data, cols, score_function, alpha=alpha, mul=mul)
+                            cache[vj][par_sea] = local_score(data, cols, score_function)
                         time_score += time.time() - start_time
                         diff_temp = cache[vj][par_sea] - cache[vj][tuple(dag[vj]['par'])]
                         if debug:
@@ -344,12 +342,11 @@ def hc(data, method='standard', score_function='default', alpha=0.3, mul=1, debu
                         par_sea_par_vi = tuple(sorted(dag[par_vi]['par'] + [vi]))
                         if par_sea_par_vi not in cache[par_vi]:
                             cols = [varnames.index(x) for x in (par_vi,) + par_sea_par_vi]
-                            cache[par_vi][par_sea_par_vi] = local_score(data, cols, score_function, alpha=alpha,
-                                                                        mul=mul)
+                            cache[par_vi][par_sea_par_vi] = local_score(data, cols, score_function)
                         par_sea_vi = tuple([x for x in dag[vi]['par'] if x != par_vi])
                         if par_sea_vi not in cache[vi]:
                             cols = [varnames.index(x) for x in (vi,) + par_sea_vi]
-                            cache[vi][par_sea_vi] = local_score(data, cols, score_function, alpha=alpha, mul=mul)
+                            cache[vi][par_sea_vi] = local_score(data, cols, score_function)
                         time_score += time.time() - start_time
                         diff_temp = cache[par_vi][par_sea_par_vi] + cache[vi][par_sea_vi] - cache[par_vi][
                             tuple(dag[par_vi]['par'])] - cache[vi][tuple(dag[vi]['par'])]
@@ -365,7 +362,7 @@ def hc(data, method='standard', score_function='default', alpha=0.3, mul=1, debu
                     par_sea = tuple([x for x in dag[vi]['par'] if x != par_vi])
                     if par_sea not in cache[vi]:
                         cols = [varnames.index(x) for x in (vi,) + par_sea]
-                        cache[vi][par_sea] = local_score(data, cols, score_function, alpha=alpha, mul=mul)
+                        cache[vi][par_sea] = local_score(data, cols, score_function)
                     time_score += time.time() - start_time
                     diff_temp = cache[vi][par_sea] - cache[vi][tuple(dag[vi]['par'])]
                     if diff_temp - diff > 1e-10:
@@ -382,8 +379,118 @@ def hc(data, method='standard', score_function='default', alpha=0.3, mul=1, debu
             print('compute score cost: %.2f%%' % (time_score / time_total * 100))
             print('others cost: %.2f%%' % (
                     (time_total - time_preprocess - time_checkcycle - time_score) / time_total * 100))
+    elif method == 'nal_ipw':
+        variety = 'aipw'
+        # calculate the run time
+        time_total = time.time()
+        time_checkcycle = 0
+        time_score = 0
+        start_time = time.time()
+        varnames = data.columns.tolist()
+        cause_list = find_causes(data)
+        cache_weight = {}
+        cache_data = {}
+        cache_score = {var: {} for var in varnames}
+
+        if all(data[var].dtype.name == 'category' for var in data):
+            data = data.apply(lambda x: x.cat.codes).to_numpy()
+        elif all(data[var].dtype.name != 'category' for var in data):
+            data = data.to_numpy()
+        else:
+            raise Exception('Mixed data is not supported.')
+        # initialise dag, cached scores and candidate parents for each variable
+        dag = {var: {'par': [], 'nei': []} for var in varnames}
+        for var in varnames:
+            cache_data, cache_weight, W = pairwise(data, varnames, [var], cause_list, cache_data, cache_weight, variety)
+            cache_score[var][tuple([])] = local_score(cache_data[W], [varnames.index(var)], 'nal_ipw', cache_weight[W])
+        candidate = {var: [v for v in varnames if v is not var] for var in varnames}
+        time_preprocess = time.time() - start_time
+        diff = 1
+        while diff > 0:
+            diff = 0
+            edge_candidate = []
+            for vi in varnames:
+                # attempt to add edges vi->vj
+                for vj in candidate[vi]:
+                    # calculate the time for checking cycles
+                    start_time = time.time()
+                    cyc_flag = check_cycle(vi, vj, dag)
+                    time_checkcycle += time.time() - start_time
+                    if not cyc_flag:
+                        # calculate the time for computing the score
+                        start_time = time.time()
+                        par_sea = tuple(sorted(dag[vj]['par'] + [vi]))
+                        if par_sea not in cache_score[vj]:
+                            cache_data, cache_weight, W = pairwise(data, varnames, [vj] + list(par_sea), cause_list,
+                                                                   cache_data, cache_weight, variety)
+                            cols = [varnames.index(x) for x in (vj,) + par_sea]
+                            cache_score[vj][par_sea] = local_score(cache_data[W], cols, 'nal_ipw', cache_weight[W])
+                        time_score += time.time() - start_time
+                        diff_temp = cache_score[vj][par_sea] - cache_score[vj][tuple(dag[vj]['par'])]
+                        if debug:
+                            print(vi, vj, diff_temp, 'a')
+                        if diff_temp - diff > 1e-10:
+                            diff = diff_temp
+                            edge_candidate = [vi, vj, 'a']
+                for par_vi in dag[vi]['par']:
+                    # attempt to reverse edges from vi<-par_vi to vi->par_vi
+                    # calculate the time for checking cycles
+                    start_time = time.time()
+                    cyc_flag = check_cycle(vi, par_vi, dag)
+                    time_checkcycle += time.time() - start_time
+                    if not cyc_flag:
+                        # calculate the time for computing the score
+                        start_time = time.time()
+                        par_sea_par_vi = tuple(sorted(dag[par_vi]['par'] + [vi]))
+                        if par_sea_par_vi not in cache_score[par_vi]:
+                            cache_data, cache_weight, W = pairwise(data, varnames, [par_vi] + list(par_sea_par_vi),
+                                                                   cause_list, cache_data, cache_weight, variety)
+                            cols = [varnames.index(x) for x in (par_vi,) + par_sea_par_vi]
+                            cache_score[par_vi][par_sea_par_vi] = local_score(cache_data[W], cols, 'nal_ipw',
+                                                                              cache_weight[W])
+                        par_sea_vi = tuple([x for x in dag[vi]['par'] if x != par_vi])
+                        if par_sea_vi not in cache_score[vi]:
+                            cache_data, cache_weight, W = pairwise(data, varnames, [vi] + list(par_sea_vi), cause_list,
+                                                                   cache_data, cache_weight, variety)
+                            cols = [varnames.index(x) for x in (vi,) + par_sea_vi]
+                            cache_score[vi][par_sea_vi] = local_score(cache_data[W], cols, 'nal_ipw', cache_weight[W])
+                        time_score += time.time() - start_time
+                        diff_temp = cache_score[par_vi][par_sea_par_vi] + cache_score[vi][par_sea_vi] - \
+                                    cache_score[par_vi][tuple(dag[par_vi]['par'])] - cache_score[vi][
+                                        tuple(dag[vi]['par'])]
+                        if diff_temp - diff > 1e-10:
+                            diff = diff_temp
+                            edge_candidate = [vi, par_vi, 'r']
+                        if debug:
+                            print(vi, par_vi, diff_temp, 'r')
+
+                    # attempt to delete edges vi<-par_vi
+                    # calculate the time for computing the score
+                    start_time = time.time()
+                    par_sea = tuple([x for x in dag[vi]['par'] if x != par_vi])
+                    if par_sea not in cache_score[vi]:
+                        cache_data, cache_weight, W = pairwise(data, varnames, [vi] + list(par_sea), cause_list,
+                                                               cache_data, cache_weight, variety)
+                        cols = [varnames.index(x) for x in (vi,) + par_sea]
+                        cache_score[vi][par_sea] = local_score(cache_data[W], cols, 'nal_ipw', cache_weight[W])
+                    time_score += time.time() - start_time
+                    diff_temp = cache_score[vi][par_sea] - cache_score[vi][tuple(dag[vi]['par'])]
+                    if diff_temp - diff > 1e-10:
+                        diff = diff_temp
+                        edge_candidate = [par_vi, vi, 'd']
+                    if debug:
+                        print(par_vi, vi, diff_temp, 'd')
+            dag, candidate = dag_update(dag, edge_candidate, candidate, diff, debug)
+        time_total = time.time() - time_total
+        if debug:
+            print('total cost:', time_total, 'seconds')
+            print('preprocess cost: %.2f%%' % (time_preprocess / time_total * 100))
+            print('check cycle cost: %.2f%%' % (time_checkcycle / time_total * 100))
+            print('compute score cost: %.2f%%' % (time_score / time_total * 100))
+            print('others cost: %.2f%%' % (
+                    (time_total - time_preprocess - time_checkcycle - time_score) / time_total * 100))
     else:
-        raise Exception('The input method:' + method + 'is invalid.')
+        raise Exception('The input method: ' + method + ' is invalid.')
     return post_process(dag, dc)
 
 
@@ -454,7 +561,7 @@ def pc_stable(data, test='default', threshold=0.05, debug=False):
 #     return post_process(dag, dc)
 
 
-def ges(data, method='complete', score_function='default', prune=True, thres=0.1, alpha=0.3, debug=False):
+def ges(data, method='complete', score_function='default', prune=True, thres=0.1, debug=False):
     '''
     greedy equivalence search algorithm
     :param data: the training data used for learn BN
@@ -658,13 +765,13 @@ def ges(data, method='complete', score_function='default', prune=True, thres=0.1
                             if parents_new not in cache_score[y][W]:
                                 cols = [varnames.index(v) for v in (y,) + parents_new]
                                 cache_score[y][W][parents_new] = local_score(cache_data[W], cols, score_function,
-                                                                             weights=cache_weights[W], alpha=alpha)
+                                                                             weights=cache_weights[W])
                             score_new = cache_score[y][W][parents_new]
 
                             if parents_cur not in cache_score[y][W]:
                                 cols = [varnames.index(v) for v in (y,) + parents_cur]
                                 cache_score[y][W][parents_cur] = local_score(cache_data[W], cols, score_function,
-                                                                             weights=cache_weights[W], alpha=alpha)
+                                                                             weights=cache_weights[W])
                             score_cur = cache_score[y][W][parents_cur]
                             diff_temp = score_new - score_cur
                             if diff_temp > diff:
@@ -712,12 +819,12 @@ def ges(data, method='complete', score_function='default', prune=True, thres=0.1
                             if parents_new not in cache_score[y][W]:
                                 cols = [varnames.index(v) for v in (y,) + parents_new]
                                 cache_score[y][W][parents_new] = local_score(cache_data[W], cols, score_function,
-                                                                             weights=cache_weights[W], alpha=alpha)
+                                                                             weights=cache_weights[W])
                             score_new = cache_score[y][W][parents_new]
                             if parents_cur not in cache_score[y][W]:
                                 cols = [varnames.index(v) for v in (y,) + parents_cur]
                                 cache_score[y][W][parents_cur] = local_score(cache_data[W], cols, score_function,
-                                                                             weights=cache_weights[W], alpha=alpha)
+                                                                             weights=cache_weights[W])
                             score_cur = cache_score[y][W][parents_cur]
                             diff_temp = score_new - score_cur
                             if diff_temp > diff:
@@ -740,6 +847,8 @@ def ges(data, method='complete', score_function='default', prune=True, thres=0.1
                 cpdag = dag2cpdag(cpdag)
             else:
                 break
+    else:
+        raise Exception('The input method: ' + method + ' is invalid.')
     time_total = time.time() - time_total
     if debug:
         print('total cost:', time_total, 'seconds')
